#pragma once

#include "value_restriction.hpp"

namespace gz {
    template<typename T>
    class RestrictedValue {
        public:
            /**
             * @brief Create a value that is restricted to the restriction
             * @throws InvalidArgument if initial does not satisfy the restriction
             */
            RestrictedValue(T&& initial, ValueRestriction<T>&& restriction);
            
            /**
             * @brief Get the value.
             * @details
             *  This function will never return a value that does not satisfy the restriction
             */
			inline const T& get() const { return val; }
            /**
             * @brief Set a new value
             * @throws InvalidArgument if initial does not satisfy the restriction
             */
			void operator=(const T& v);
			void operator=(T&& v);
            void operator=(const std::string& s) const requires (ConstructibleFromString<T> && !std::same_as<std::string, T>);
			inline const ValueRestriction<T>& getRestriction() const { return restriction; }

        private:
            T val;
            ValueRestriction<T> restriction;
    };
}

namespace gz {
	template<typename T>
	RestrictedValue<T>::RestrictedValue(T&& initial, ValueRestriction<T>&& restriction) {
		
	}

	template<typename T>
	void RestrictedValue<T>::operator=(T&& v) {
		if (restriction.isValueAllowed(v)) {
			val = std::move(v);
		}
		else {
			throw InvalidArgument("Trying to assign invalid value", "RestrictedValue::operator=");
		}
	}
	template<typename T>
	void RestrictedValue<T>::operator=(const T& v) {
		if (restriction.isValueAllowed(v)) {
			val = v;
		}
		else {
			throw InvalidArgument("Trying to assign invalid value", "RestrictedValue::operator=");
		}
	}
	template<typename T>
	void RestrictedValue<T>::operator=(const std::string& s) const requires (ConstructibleFromString<T> && !std::same_as<std::string, T>) {
		T v;
		try {
			v = fromString<T>(s);
		}
		catch (...) {
			throw InvalidArgument(std::string("Can not convert string to type <") + typeid(T).name() + ">", "RestrictedValue::operator=");
		}
		*this = std::move(v);
    }

    
}  // namespace gz
