#pragma once

#include "common.hpp"

#include <functional>
#include <type_traits>

namespace gz {
    enum SettingsManagerAllowedValueTypes {
        SM_RANGE, SM_LIST,
    };

    /**
     * @brief A value restriction can check if a variable satisfies the restriction.
     * @details
     *  The restriction can be:
     *  - SM_RANGE -> value needs to be in [min, max]
     *  - SM_LIST -> value must be one of
     */
    template<typename T>
    class ValueRestriction {
        public:
            /**
             * @todo allow any class that has <= and >=
             * @brief Restriction is a range [lower, upper]
             */
            ValueRestriction(const T& lower, const T& upper) requires util::Number<T>;
            /**
             * @brief Restriction is a list of allowedValues
             */
            ValueRestriction(std::vector<T>&& allowedValues) noexcept;

            /**
             * @brief Check if a value is allowed according to SettingsManagerAllowedValues
             */
            bool isValueAllowed(const T& value) const noexcept;

            /**
             * @brief Construct a value from string and check if it is allowed according to SettingsManagerAllowedValues
             * @throws Exceptions that occur while s is converted to T (using gz::fromString())
             */
            bool isValueAllowed(const std::string& s) const requires (ConstructibleFromString<T> && !std::same_as<std::string, T>);

			inline const SettingsManagerAllowedValueTypes& getType() const { return type; }
        private:
            const SettingsManagerAllowedValueTypes type;
            std::vector<T> allowedValues;  // using vector and not set to also store lower, upper in there
    };


}  // namespace gz

namespace gz {
    //
    // VALUERESTICTION IMPLEMENTATION
    //
	template<typename T>
	ValueRestriction<T>::ValueRestriction(const T& lower, const T& upper) requires util::Number<T> 
        : type(SettingsManagerAllowedValueTypes::SM_RANGE) {
        if (lower > upper) {
            throw InvalidArgument("ValueRestriction: lower(=\"" + gz::toString(lower) + "\") is larger than upper(=\"" + gz::toString(upper) + "\")");
        }
        allowedValues.push_back(lower);
        allowedValues.push_back(upper);
	}

	template<typename T>
	ValueRestriction<T>::ValueRestriction(std::vector<T>&& allowedValues) noexcept
        : type(SettingsManagerAllowedValueTypes::SM_LIST) {
        this->allowedValues = std::move(allowedValues);
	}

	template<typename T>
    bool ValueRestriction<T>::isValueAllowed(const std::string& s) const requires (ConstructibleFromString<T> && !std::same_as<std::string, T>) {
        T val = fromString<T>(s);
        return isValue(val);
	}

	template<typename T>
	bool ValueRestriction<T>::isValueAllowed(const T& value) const noexcept {
        switch (this->type) {
            case SM_LIST: {
                for (auto it = this->allowedValues.begin(); it != this->allowedValues.end(); it++) {
                    if (*it == value) { return true; }
                }
            }
            break;
            case SM_RANGE: {
                bool valid = true;
                // value >= lowest 
                valid &= value >= allowedValues.at(0);
                // value <= highest 
                valid &= value <= allowedValues.at(1);
                return valid;
            }
            break;
        } // switch
        return false;
    }
}  // namespace gz


#ifdef GZ_UTIL_TESTING 
// TODO
TEST_CASE("")

    CHECK_THROWS()


#endif
